mod exec;
mod rpc;

use std::{
    pin::Pin,
    sync::{Arc, Mutex},
};

use sysinfo::{CpuExt, DiskExt, System, SystemExt};
use tokio_stream::{wrappers::ReceiverStream, Stream, StreamExt};
use tonic::{transport::server::Router, Request, Response, Status};

pub struct Server {
    sys: Arc<Mutex<System>>,
}

type Result<T> = std::result::Result<T, tonic::Status>;

#[tonic::async_trait]
impl rpc::agent_server::Agent for Server {
    async fn echo(&self, req: Request<rpc::EchoRequest>) -> Result<Response<rpc::EchoResponse>> {
        Ok(Response::new(rpc::EchoResponse {
            message: req.into_inner().message,
        }))
    }

    async fn get_sys_info(&self, _: Request<()>) -> Result<Response<rpc::SysInfoResponse>> {
        let sys = self.sys.lock().unwrap();

        let cpus = sys
            .cpus()
            .iter()
            .map(|cpu| rpc::sys_info_response::Cpu {
                freq_mhz: cpu.frequency(),
                usage: cpu.cpu_usage(),
            })
            .collect();

        let disks = sys
            .disks()
            .iter()
            .map(|disk| rpc::sys_info_response::Disk {
                name: disk.name().to_string_lossy().into_owned(),
                total_bytes: disk.total_space(),
                avail_bytes: disk.available_space(),
                mount_point: disk.mount_point().to_string_lossy().into_owned(),
            })
            .collect();

        let response = Response::new(rpc::SysInfoResponse {
            uptime: sys.uptime(),
            hostname: sys.host_name().unwrap_or_default(),
            os: sys.long_os_version().unwrap_or_default(),
            mem_total_bytes: sys.total_memory(),
            mem_avail_bytes: sys.available_memory(),
            swap_total_bytes: sys.total_swap(),
            swap_free_bytes: sys.free_swap(),
            cpus,
            disks,
        });

        Ok(response)
    }

    type ExecStream = Pin<Box<dyn Stream<Item = Result<rpc::ExecResponse>> + Send + Sync>>;

    async fn exec(&self, req: Request<rpc::ExecRequest>) -> Result<Response<Self::ExecStream>> {
        use exec::*;

        let rpc::ExecRequest { program, args } = req.into_inner();

        match exec(&program, &args) {
            Ok(receiver) => {
                let stream = ReceiverStream::new(receiver).map(|inner| {
                    Ok(rpc::ExecResponse {
                        response: Some(inner.into()),
                    })
                });

                Ok(Response::new(Box::pin(stream)))
            }
            Err(err) => Err(Status::failed_precondition(err.to_string())),
        }
    }
}

pub fn new_server() -> Router {
    let server = Server {
        sys: Arc::new(Mutex::new(System::new_all())),
    };

    tonic::transport::Server::builder().add_service(rpc::agent_server::AgentServer::new(server))
}

#[cfg(test)]
mod tests {
    use std::net::SocketAddr;

    use tokio::net::TcpListener;
    use tokio_stream::wrappers::TcpListenerStream;

    use super::*;

    async fn spawn_server() -> SocketAddr {
        let listener = TcpListener::bind("[::]:0").await.unwrap();
        let address = listener.local_addr().unwrap();

        let listener = TcpListenerStream::new(listener);

        tokio::spawn(async move { new_server().serve_with_incoming(listener).await.unwrap() });

        address
    }

    #[tokio::test]
    async fn echo_works() {
        let addr = spawn_server().await;

        let mut client =
            rpc::agent_client::AgentClient::connect(format!("http://[::]:{}", addr.port()))
                .await
                .unwrap();

        let message = "Hello!".to_owned();
        let response = client
            .echo(rpc::EchoRequest {
                message: message.clone(),
            })
            .await
            .expect("to respond");

        assert_eq!(rpc::EchoResponse { message }, response.into_inner());
    }
}
